"""
The authentication module is responsible for taking log in credentials an establish a session.
Session data is saved in the Session collection.
"""
from abc import ABCMeta

from of.broker.lib.schema_mongodb import mbe_object_id

__author__ = "Nicklas Boerjesson"

import hashlib



import datetime

from decorator import decorator
from of.common.aspect_utilities import alter_function_parameter_and_call
from inspect import getfullargspec

_authentication = None


def init_authentication(_authentication_backend):
    """
    Initializes the global authentication object

    :param _authentication_backend: An instance of an AuthenticationBackend subclass
    :return: an instance of the global authentication object
    """
    global _authentication
    _authentication = Authentication(_authentication_backend)
    return _authentication


def check_session(_session_id):
    """
       Global convenience function, see Authentication.check_session

       :param _session_id: The session Id.

    """
    global _authentication
    if _authentication is None:
        raise AuthenticationError('check_session: No authentication object initialized. Call init_authentication.')

    # TODO: ITRT we probably have to make this a proxy to thread-local instances.

    # Check if the session is valid.
    return _authentication.check_session(_session_id)


def login(_credentials):
    """
       Global convenience function, see Authentication.logout

       :param _credentials: An MBE credentials package.

    """
    global _authentication
    if _authentication is None:
        raise AuthenticationError('password_login: No authentication object initialized. Call init_authentication.')

    # Check if the session is valid.
    return _authentication.login(_credentials)


def logout(_session_id):
    """
       Global convenience function, see Authentication.logout

       :param _session_id: The session Id

    """
    global _authentication
    if _authentication is None:
        raise AuthenticationError('session logout: No authentication object initialized. Call init_authentication.')

    # Check if the session is valid.
    return _authentication.logout(_session_id)


def check_session_aspect(func, *args, **kwargs):
    """
    The authentication aspect code, provides an aspect for the aop_check_session decorator

    """

    # Get a list of the names of the non-keyword arguments
    _argument_specifications = getfullargspec(func)
    try:
        # Try and find _session_id
        arg_idx = _argument_specifications.args.index("_session_id")
    except:
        raise Exception("Authentication aspect for \"" + func.__name__ + "\": No _session_id parameter")

    # Check if the session is valid.
    _user = check_session(args[arg_idx])

    # Call the function and set the _user parameter, if existing.
    return alter_function_parameter_and_call(func, args, kwargs, '_user', _user)


def aop_check_session(f):
    """
        A session validation decorator, searches that parameters for _session_id and checks that session
    """
    return decorator(check_session_aspect, f)


class AuthenticationError(Exception):
    """Exception class for authentication errors"""
    pass


class AuthenticationBackend(metaclass=ABCMeta):

    def get_session(self, session_id):
        raise Exception("AuthenticationBackend.get_session is not implemented!")

    def get_user(self, user_id):
        raise Exception("AuthenticationBackend.get_user is not implemented!")

    def authenticate_username_password(self, _credentials):
        raise Exception("AuthenticationBackend.authenticate_username_password is not implemented!")

    def logout(self, _session_id):
        raise Exception("AuthenticationBackend.logout is not implemented!")

class Authentication():
    """
    The authentication class is the central entity for authentication in IF.
    It handles login, logout and session validation.
    """

    authentication_backend = None

    def __init__(self, _authentication_backend):
        """
        The Authentication class need access to the database for user information.
        :param _authentication_backend: An instance of an AuthenticationBackend subclass
        """
        self.authentication_backend = _authentication_backend

    def check_session(self, _session_id):
        """
        Checks if a session_id is valid and returns the user object
        Otherwise raises an AuthenticationError,

        :param _session_id: A session_id, format is a MongoDB objectId.
        :return: A user object

        """

        _session = self.authentication_backend.get_session(_session_id)
        if _session is not None:
            _user = self.authentication_backend.get_user(_session["user_id"])

            if _user is not None:
                return _user
            else:
                raise AuthenticationError("Authentication failed. The user associated with the session"
                                          " has been removed.\nSessionId: " + str(_session_id))


        else:
            raise AuthenticationError("Authentication failed for sessionId : " + str(_session_id))

    def login(self, _credentials):
        """
        Takes a credentials object and logs in. Returns a tuple with a session_id and a user object if successful.
        Otherwise raises an AuthenticationError,
        If user is already logged in, that session is destroyed and this one created.

        :param _credentials: A structure containing credentials
        :return: A tuple with a session_id and a user object

        """

        if "usernamePassword" in _credentials:
            _session_id, _user = self.authentication_backend.authenticate_username_password(_credentials)

            if _session_id is None:
                raise AuthenticationError("login error: Invalid login/username combination.")
            else:
                return _session_id, _user

        else:
            if len(_credentials) > 0:
                raise AuthenticationError("login error: Invalid authentication scheme \"" + _credentials[0] + "\"")
            else:
                raise AuthenticationError("login error: Invalid authentication structure (empty)")

    def logout(self, _session_id):
        """
        Logs out the session

        :param _session_id: The session Id
        :return:

        """
        self.authentication_backend.logout(_session_id)


"""
    A test server in CherryPy that demonstrates how to use MBE from CherryPy.
"""
from of.broker.lib.auth_backend import MongoDBAuthBackend
from of.common.security.authentication import init_authentication
from of.schemas.validation import of_uri_handler

__author__ = 'Nicklas Boerjesson'

import sys
import os

# Add MBE path
script_dir = os.path.dirname(__file__)
sys.path.append(os.path.join(script_dir, "../"))

import cherrypy
from of.broker.cherrypy_api.node import CherryPyNode
from of.broker.testing.init import init_database


class Server(object):
    """
        The server object.
    """
    node = None
    database_access = None

    def __init__(self):
        self._reset_database()
        init_authentication(MongoDBAuthBackend(self._database_access))
        self.node = CherryPyNode(self._database_access)

    def _reset_database(self):
        self._database_access = init_database("test_skeleton_of",
                                              _data_files=[os.path.join(script_dir, "../../testing/data_struct.json")],
                                              _json_schema_folders=[os.path.join(script_dir, "../../../schemas")],
                                              _uri_handlers={"of": of_uri_handler})

    @cherrypy.expose
    @cherrypy.tools.json_out(content_type='application/json')
    def reset_database(self, **kwargs):
        self._reset_database()
        return {"result": "OK"}


def start_server():
    """
        Initialize and start the server.
    """
    if hasattr(cherrypy.engine, 'subscribe'):  # CherryPy >= 3.1
        pass
    else:
        raise Exception("This application requires CherryPy >= 3.1 or higher.")
        # cherrypy.engine.on_stop_engine_list.append(_save_data)

    # Some global configuration

    cherrypy.config.update({"server.ssl_certificate": os.path.join(os.path.dirname(__file__), "../cert.pem"),
                            "server.ssl_private_key": os.path.join(os.path.dirname(__file__), "../privkey.pem")})

    # cherrypy.config.update(config=os.path.join(script_dir, "config.ini"))
    app = cherrypy.tree.mount(Server(), '/',
                              {'/':
                                  {
                                      "tools.staticdir.root": os.path.abspath(
                                          os.path.join(os.path.dirname(__file__), "static")),
                                      "tools.decode.on": True,
                                      "tools.trailing_slash.on": True,
                                      "tools.staticdir.on": True,
                                      "tools.staticdir.index": "index.html",
                                      "tools.staticdir.dir": ""
                                  }
                              })

    if hasattr(cherrypy.engine, "signal_handler"):
        cherrypy.engine.signal_handler.subscribe()
    if hasattr(cherrypy.engine, "console_control_handler"):
        cherrypy.engine.console_control_handler.subscribe()
    cherrypy.engine.start()
    cherrypy.engine.block()


if __name__ == '__main__':
    start_server()
else:
    cherrypy.config.update({"server.ssl_certificate": os.path.join(os.path.dirname(__file__), "../cert.pem"),
                            "server.ssl_private_key": os.path.join(os.path.dirname(__file__), "../privkey.pem"),
                            "tools.staticdir.root": os.path.abspath(script_dir)})
    app = cherrypy.tree.mount(Server())
    print("Running")



"""
     The access module contains features providing an abstraction layer over MongoDB.
"""

import copy

from bson import ObjectId
from jsonschema.exceptions import ValidationError

from of.schemas.schema import SchemaTools
from of.broker.lib.logging import Logging

__author__ = 'nibo'


class DatabaseAccess():
    """
        The database access class handles all communication with the database
        It provides logging, validation against a schema and hiding ObjectIds.
    """

    database = None
    schema_tools = None
    logging = None

    def __init__(self, _database, _json_schema_folders=None, _uri_handlers = None, _schema_tools = None):
        """
        Initialize a database access object
        :param _database: A MongoDB client object
        :param _json_schema_folders: A list of application specific JSON schema folders

        """

        self.database = _database

        self.logging = Logging(_database=self.database)

        # Database access is dependent on schema tools for validation.
        if _schema_tools:
            self.schema_tools = _schema_tools
        else:
            self.schema_tools = SchemaTools(_json_schema_folders = _json_schema_folders, _uri_handlers=_uri_handlers)

    def verify_document(self, _data, _collection_name, _caller_name):
        """
        Check so that the collection for the specified schema in the input data match the supplied collection in
        _collection_name.

        :param _data: The data to check.
        :param _collection_name: Name of the collection that it should be checked against
        :param _caller_name: The name of the calling function for logging


        """

        if "schemaRef" in _data:
            try:
                _curr_schema = self.schema_tools.json_schema_objects[_data["schemaRef"]]
            except:
                raise Exception(
                    _caller_name + ": Error - invalid data structure, schemaRef not found: " + _data["schemaRef"])

            if _curr_schema["collection"] != _collection_name:
                raise Exception(
                    _caller_name + ": Invalid schemaRef: \"" + _curr_schema["collection"] +"\". " + _caller_name + " is restricted to the \"" +
                    _collection_name + "\"-collection.")
        else:
            # An MBE document must *always* have a schemaRef property
            raise Exception(_caller_name + ": Missing schemaRef.")

    @staticmethod
    def verify_condition(_data, _collection_name, _caller_name):
        """
        Validates the conditions' structure and the collection it queries.

        :param _data: The data to check.
        :param _collection_name: Name of the collection that it should be checked against
        :param _caller_name: The name of the calling function for logging

        """

        if "conditions" not in _data:
            raise Exception(_caller_name + ": Error - Not a valid condition, \"conditions\"-property missing")

        if "collection" in _data and _data["collection"] != _collection_name:
            raise Exception(_caller_name + ": Error - collection omitted or wrong, only queries against"
                                           " the " + _collection_name + " collection is permitted.")

    def manage_input(self, _input, _schema_ref=None):
        """
        Validate, convert _id to objectId instances and parse collection from input, whether it is data to save or a\
        condition.
        Note: This should not be run twice on the same data.

        :param _input: Data to handle
        :param _schema_ref: If set the schema to validate against
        :return: A tuple with the data and the database collection specified in the schema.

        """
        try:
            # Apply(if _schema_ref is set, validate against that schema instead)
            _input, _schema_obj = self.schema_tools.apply(_input, _schema_ref)
        except ValidationError as e:
            raise ValidationError("handle_input: Validation error:" + str(e) + ". Data: \n" + str(_input))
        except Exception as e:
            raise Exception(
                "handle_input: Non-validation error from validation: Type: " + e.__class__.__name__ + " Message:" + str(
                    e) + ". Data: \n" + str(_input))

        # In a condition, the input data is under the conditions-field
        if _schema_ref == "of://conditions.json":
            _data = copy.deepcopy(_input["conditions"])
            _collection = _input["collection"]
        else:
            _collection = _schema_obj["collection"]
            _data = copy.deepcopy(_input)

        return _data, self.database[_collection]

    @staticmethod
    def load_exactly_one_document(_collection, _id, _zero_error=None, _duplicate_error=None):
        """
        Raises provided errors if not exactly one row in result.

        :param _collection: The collection to load from
        :param _id: The _id of the document
        :param _zero_error: Error to report if no documents are found
        :param _duplicate_error: Error to report when duplicate documents are found
        :return: The document
        """

        _document_cursor = _collection.find({"_id": _id})
        if _document_cursor.count() == 0:
            if _zero_error is not None:
                raise Exception(_zero_error + "_id: " + str(
                    _id) + " could not be found.")
            else:
                return None
        elif _document_cursor.count() > 1:
            if _duplicate_error is not None:
                raise Exception(
                    "Access.save: Tried saving existing, but found duplicate documents with(try to remove either) _id: "
                    + str(_id) + ".")
            else:
                return None
        elif _document_cursor.count() == 1:
            return _document_cursor[0]

    def save(self, _document, _user, _old_document=None, _allow_save_id=False):
        """
        Save a document to a collection

        :param _document: The data to save
        :param _user: A user object
        :param _old_document: If available, the existing data, used to avoid an extra read when logging changes.
        :param _allow_save_id: If set, to not assume that _id being set means updating an existing document.
        :return: The object id of the saved document

        """
        if _user is not None:
            _user_id = _user["_id"]
        else:
            _user_id = None

        _document, _collection = self.manage_input(_document)

        if _old_document is None and "_id" in _document:
            # Load the existing data
            if not _allow_save_id:
                _zero_error = "Access.save: Tried to save data over existing but didn't find an existing node."
            else:
                _zero_error = None

            _old_document = self.load_exactly_one_document(_collection, _document["_id"], _zero_error=_zero_error,
                                                           _duplicate_error="Access.save: Tried to save data over "
                                                                            "existing but found duplicate nodes")
            if (_old_document is not None) and (str(_old_document["schemaRef"]) != str(_document["schemaRef"])):
                raise Exception(
                    "Access.save: Cannot change schema of an existing node, remove and add. _id: " + str(
                        _document["_id"]) + ", new:" + str(_old_document["schemaRef"]) + ", old:" + str(
                        _old_document["schemaRef"]))

        if _old_document is not None:
            _old_document, _dummy_collection = self.manage_input(_old_document)

        _result = str(_collection.save(_document))
        if _collection.name != "log":
            self.logging.log_save(_document, _user_id, _old_document)
        return _result

    def remove_documents(self, _documents, _user, _collection_name=None):
        """
        Remove the documents in the documents list

        :param _documents: The list of documents
        :param _user: A user object
        :param _collection_name: The collection from where to remove them.
        :return: Nothing
        """
        if _user is not None:
            _user_id = _user["_id"]
        else:
            _user_id = None
        _result = []
        # Loop through documents and remove them
        _collection = self.database[_collection_name]
        for _document in _documents:
            # noinspection PyUnusedLocal
            _result.append(_collection.remove(_document))

            self.logging.log_remove(_document, _user_id)

        return _result

    def remove_condition(self, _condition, _user):
        """
        Remove documents that match the supplied MBE condition

        :param _condition: A MongoDB search criteria
        :return:
        """
        if _user is not None:
            _user_id = _user["_id"]
        else:
            _user_id = None
        _raw_condition, _collection = self.manage_input(_condition, "of://conditions.json")
        _removed_documents_cursor = _collection.find(_raw_condition)

        _documents = [x for x in _removed_documents_cursor]
        _result = []
        for _document in _documents:
            # noinspection PyUnusedLocal
            _result.append(_collection.remove(_raw_condition))
            self.logging.log_remove(_document, _user_id)

        return _result

    def find(self, _conditions, _do_not_fix_object_ids=False):
        """
        Return a list of documents that match the supplied MBE condition.
        :param _conditions: An MBE condition
        :return: A list of matching documents
        """

        def _recurse_object_ids(_data):

            if isinstance(_data, list):
                _destination = []
                for _curr_row in _data:
                    _destination.append(_recurse_object_ids(_curr_row))

                return _destination
            elif isinstance(_data, dict):
                _destination = {}
                for _curr_key, _curr_value in _data.items():
                    _destination[_curr_key] = _recurse_object_ids(_curr_value)
                return _destination

            elif isinstance(_data, ObjectId):
                return str(_data)
            else:
                return _data

        _raw_conditions, _collection = self.manage_input(_conditions, "of://conditions.json")
        _result = list(_collection.find(_raw_conditions))
        if not _do_not_fix_object_ids:
            return _recurse_object_ids(_result)
        else:
            return _result

    # noinspection PyMethodMayBeStatic
    def transform_collection(self, _destination_schema, _map, _row_callback):
        """
        NOT IMPLEMENTED.
        :param _destination_schema:
        :param _map:
        :param _row_callback:
        :return:
        """

        pass
        # Iterate all documents

        # Copy _id to new document
        # Copy all data using map and field id:s to document
        # If set, use callback to finish what has to be manually done


        # Write cumulative log entry using map changes and results from _row_callback


"""
This module exposes the Optimal BPM Broker API as a web service through a CherryPy module
Note that most of its initialisation is performed in the startup ../broker.py
"""

import copy
import time
from multiprocessing import Queue

import cherrypy

from of.common.logging import EC_SERVICE, write_to_log, SEV_DEBUG, EC_NOTIFICATION, EC_PROBE, SEV_WARNING, \
    severity_identifiers, category_identifiers
from of.common.messaging.constants import UNEXPECTED_CONDITION
from of.broker.cherrypy_api.node import aop_login_json, aop_check_session, CherryPyNode
from of.schemas.constants import id_right_admin_everything
from of.common.security.groups import has_right, aop_has_right
from of.broker.lib.node import sanitize_node
from of.common.messaging.utils import get_environment_data
from of.schemas.constants import peer_type_to_schema_id


__author__ = 'Nicklas Borjesson'


class CherryPyBroker(object):
    """
    The root web server class of the Optimal BPM Broker.
    It exposes the web services of the broker.
    Note: The web application admin UI is served statically from the /admin folder.
    """

    #: A sessionId-indexed dictionary of logged in peers
    peers = None


    #: Plugin management
    plugins = None

    #: A reference to a Nodes instance
    node = None

    def __init__(self, _process_id, _address, _database_access):
        """
        Initializes the broker web service and includes and initiates the other parts of the API as well
        :param _database_access: A DatabaseAccess instance for database connectivity
        :param _process_id: The system process id of the broker
        :param _address: The peer address of the broker
        :param _stop_broker: A callback to a function that shuts down the broker
        """
        write_to_log(_process_id=_process_id, _category=EC_SERVICE, _severity=SEV_DEBUG,
                     _data="Initializing broker REST API.")
        self.peers = {}
        self.process_id = _process_id
        self.address = _address

        self.node = CherryPyNode(_database_access=_database_access)

    def write_debug_info(self, _data):
        write_to_log(_data=_data, _category=EC_NOTIFICATION, _severity=SEV_DEBUG, _process_id=self.process_id)

    @cherrypy.expose
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    @aop_has_right([id_right_admin_everything])
    def get_broker_environment(self, **kwargs):
        self.write_debug_info("Request for broker information")
        return get_environment_data()

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_login_json
    def register(self, **kwargs):
        """
        Register a peer with the web service.
        :param kwargs: A structure containing credentials, environment data, peer type and peer address.
        :return: A structure containing the sessionId and settings of the peer.
        """
        # Note: The input to this function is not validated by against the register.json schema and should not be,
        # as it is both a simple structure and would be a potential DOS-voulnerability. All other calls require a valid
        # session cookie before even entering any logic, and are therefore more protected that this function.
        self.write_debug_info("Register called")

        try:
            _data = kwargs["_message"]
            _peer_type = _data["peerType"]
        except KeyError as e:
            write_to_log("Register: A peer at " + str(cherrypy.request.remote.ip) +
                         " tried to register without the " + str(e) + " key in the dict.",
                         _category=EC_PROBE, _severity=SEV_WARNING)
            return None

        try:

            if _data["address"] is None:
                if _peer_type == "admin":
                    # Generate an address, like "admin_root".
                    _address = _peer_type + "_" + kwargs["_user"]["name"]
                else:
                    write_to_log("The " + _peer_type + " peer type must state an address.",
                                 _category=EC_PROBE, _severity=SEV_WARNING)
                    return None
            else:
                _address = _data["address"]

            # TODO: Implement a "defaults" node so that we can return a default empty node here. (PROD-41)

            # Load the matching node for the address
            _condition = {"schemaRef": peer_type_to_schema_id(_peer_type), "address": _address}
            _settings = sanitize_node(self.admin.node._node.find(_condition, kwargs["_user"]))
            _session_id = kwargs["_session_id"]
            self.write_debug_info("New _session_id : " + str(_session_id))

            # Log out any old sessions
            for _curr_session_id, _curr_peer in dict(self.peers).items():
                if _curr_peer["address"] == _address and _curr_session_id != _session_id:
                    self.write_debug_info(
                        "Removing old registration for the peer at " + _address + ": " + _curr_session_id)
                    if "websocket" in _curr_peer:
                        try:
                            self.write_debug_info("Close remaining websocket: " + _address + ": " + _curr_session_id)
                            _curr_peer["websocket"].close(code=UNEXPECTED_CONDITION, reason='Peer logging in again')
                        except Exception as e:
                            self.write_debug_info("Exception doing so, ignoring error: " + str(e))
                    del self.peers[_curr_session_id]

            self.peers[_session_id] = {
                "user": kwargs["_user"],
                "ip": str(cherrypy.request.remote.ip),
                "address": _address,
                "environment": _data["environment"],
                "type": _peer_type,
                "queue": Queue()
            }

            self.write_debug_info("Register: A peer at " + str(
                cherrypy.request.remote.ip) + " registered with this data:" + str(_data))
            return {"session_id": _session_id, "settings": _settings}

        except Exception as e:
            # Wait to take edge off attacks
            time.sleep(3)
            # TODO: Create stats on attempts to block attacking IP:s (PROD-133)

            # report all other errors as probes as it isn't humans making typing errors
            write_to_log("Register: A peer at " + str(
                cherrypy.request.remote.ip) + " failed logging in with this data:" + str(_data) + "\nError:" + str(
                e), _category=EC_PROBE, _severity=SEV_WARNING)

            return None

    @cherrypy.expose
    def socket(self):
        """
        Called when a client wants to upgrade to a websocket. Currently only implemented for logging purposes.
        """
        self.write_debug_info("Broker: Got a /socket upgrade web socket request.")

    @cherrypy.expose
    def status(self):
        """
        Method for checking if the broker is up. No session id required and no checks performed.
        :return: The string "up" if the broker is up
        """
        return "up"

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @aop_check_session
    def write_to_log(self, **kwargs):
        """
        This function provides a way for registered peers to write to the log even if they have no web socket.
        :param kwargs:
        :return:
        """
        _session_id = kwargs["_session_id"]
        # Only registered peers get to write to the log, other trying is a probe
        if _session_id not in self.peers:
            write_to_log("Client that wasn't registered as peer tried to write to log ",
                         _category=EC_PROBE, _severity=SEV_WARNING)
            return None

        _message = cherrypy.request.json
        write_to_log(_data=_message.get("data"),
                     _category=category_identifiers.index(_message.get("category")),
                     _severity=severity_identifiers.index(_message.get("severity")),
                     _occurred_when=_message.get("occurred_when"),
                     _address=_message.get("address"),
                     _process_id=_message.get("process_id"),
                     _user_id=_message.get("user_id"),
                     _pid=_message.get("pid"),
                     _uid=_message.get("uid"),
                     _node_id=_message.get("node_id")
                     )
        return "{}"



"""
This module implements the OptimalWebSocket class and maintains the monitor global variable.
"""
import os
import threading
from queue import Empty
import json
import traceback


from of.common.logging import write_to_log, EC_COMMUNICATION, SEV_DEBUG, SEV_ERROR, EC_INTERNAL, SEV_INFO, \
    make_sparse_log_message
import of.common.logging
from of.common.messaging.factory import reply_with_error_message
from of.common.messaging.constants import ABNORMAL_CLOSE

from ws4py.messaging import TextMessage

__author__ = 'Nicklas Borjesson'

"""
The monitor global variable is used to be able to register the web socket with the handler.
This is not optimal, but the ws4py offer no other way to reach outside its contexts, as it has no parameters.
"""
monitor = None


class OptimalWebSocket(object):
    """
    This is a parent class that is use by both client- and server-side web sockets to provide Optimal Framework-specifics.
    It monitors a message_queue, and transmits information appearing in that queue to another connected web socket.
    It also handles session information and registering with the central message monitor.
    """

    #: The peer name of the client the web socket is connected to
    address = None

    #: The peer name of the client the web socket is located on
    address_own = None

    #: The session id of the socket
    session_id = None

    #: The optimal framework process id
    process_id = None
    #: The thread responsible for monitoring the queue of messages to be sent.
    monitor_message_queue_thread = None

    #: The message queue of the socket
    message_queue = None

    # The socket is connected
    connected = None

    #: A string containing the classname and address classname(address) for use with logging
    log_prefix = None

    def init(self, _session_id):
        """
        Initialize the socket, set the session id an register the web socket with the handler, start monitoring queue.
        :param _session_id: The sessionid
        """
        self.session_id = _session_id
        self.log_prefix = str(os.getpid()) + "-" + self.__class__.__name__ + "(No address yet): "
        self.connected = False
        try:
            monitor.handler.register_web_socket(self)
        except Exception as e:
            raise Exception(write_to_log("OptimalWebSocket: Exception registering web socket:" + str(e),
                                         _category=EC_COMMUNICATION, _severity=SEV_ERROR))

        # Generate log prefix string for performance
        self.log_prefix = str(os.getpid()) + "-" + self.__class__.__name__ + "(" + str(self.address) + "): "

        self.start_monitoring_message_queue()

    def write_dbg_info(self, _data):
        write_to_log(self.log_prefix + str(_data),
                     _category=EC_COMMUNICATION, _severity=SEV_DEBUG, _process_id=self.process_id)

    def start_monitoring_message_queue(self):
        """
        Start monitoring the message queue
        """
        self.monitor_message_queue_thread = threading.Thread(target=self.monitor_message_queue,
                                                             name=self.address + " message queue monitor thread")
        self.monitor_message_queue_thread.start()
        self.write_dbg_info("Message queue thread started: " + str(self.monitor_message_queue_thread.name))


    def opened(self):
        """
        Implement to have something happen when the socket is opened.
        """
        self.connected = True

    def received_message(self, message):
        """
        Receives a message from a peer and puts it on the queue

        :param message: The message
        """


        if str(message) != "":
            self.write_dbg_info("Got this message(putting on queue):" + str(message))
            if isinstance(message, TextMessage):
                monitor.queue.put([self, json.loads(str(message))])
            else:
                monitor.queue.put([self, json.loads(message)])
        else:
            self.send_message(reply_with_error_message(_runtime_instance=os.getpid(),
                                                       _error_message="Cannot send empty messages to agent",
                                                       _message={}))


    def queue_message(self, message):
        """
        Puts a message on the message queue
        :param message: The message
        :return:
        """
        self.message_queue.put(message)

    def monitor_message_queue(self):
        """
        Monitors the messages queue. Stops when the queue-threads' terminated attribute is set to True
        """
        self.monitor_message_queue_thread.terminated = False
        while not self.monitor_message_queue_thread.terminated:
            try:
                _message = self.message_queue.get(True, .1)
                try:
                    self.send_message(_message)
                except Exception as e:
                    write_to_log(self.log_prefix + "Error sending message:" + str(e) + "\nTraceback:" + traceback.format_exc(),
                                 _category=EC_COMMUNICATION, _severity=SEV_ERROR)
            except Empty:
                pass
            except Exception as e:
                write_to_log(self.log_prefix + "Error accessing send queue:" + str(e) + "\nTraceback:" + traceback.format_exc(),
                             _category=EC_INTERNAL, _severity=SEV_ERROR)

        self.write_dbg_info(" stopped message queue monitoring. Exiting thread \"" +
              str(self.monitor_message_queue_thread.name) + "\"")



    def send_message(self, message):
        """
        Sends a message to the connected counterpart web socket
        :param message: A string containing the message
        """
        if of.common.logging.severity < SEV_INFO:
            # We cannot use the normal facility here as that would cause recursion
            print(make_sparse_log_message("Sending message:" + str(message), _category=EC_COMMUNICATION,
                                          _severity = SEV_DEBUG, _address=self.address_own, _process_id=self.process_id))
        # send() below is implemented by multiple inheritance in the subclass. Ignore "unresolved attribute"-warning.
        self.send(bytes(json.dumps(message).encode()))

    def closed(self, code, reason=None):
        """
        Called when the socket is closed.
        :param code: A web socket error code as defined in: http://tools.ietf.org/html/rfc6455#section-7.4.1
        :param reason: A string describing the reason for closing the connection
        """
        # TODO: Handle the rest of the possible web socket error codes
        self.connected = False
        if code == ABNORMAL_CLOSE:
            write_to_log(self.log_prefix + "The connection to " +  self.address + " has been abnormally closed.",
                         _category=EC_COMMUNICATION, _severity=SEV_ERROR)
            self.close(code=code, reason=reason)
        else:
            self.monitor_message_queue_thread.terminated = True
            self.write_dbg_info(self.log_prefix + "Closed, code: " + str(code) + ", reason: " + str(reason))


    def error_handler(self, exception):
        """
        If an unhandled error occurs, unregister at the broker, and close the socket
        """
        write_to_log(self.log_prefix + "An exception handle in the socket, closing. Error: " + str(exception),
                                 _category=EC_COMMUNICATION, _severity=SEV_ERROR)
        monitor.handler.unregister_web_socket(self)

        # Tell the socket to close
        return False



import os
import servicemanager
import socket
import sys
import win32event
import win32service

import win32serviceutil

# The directory of the current file
script_dir = os.path.dirname(__file__)

# Add relative optimal bpm path
sys.path.append(os.path.join(script_dir, "../../"))

from of.common.win32svc import write_to_event_log
import of.broker.broker


class BrokerService(win32serviceutil.ServiceFramework):
    _svc_name_ = "Optimal BPM Broker Service"
    _svc_display_name_ = "The Optimal BPM Broker Service"

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        socket.setdefaulttimeout(60)

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)
        _exit_status = of.broker.broker.stop_broker("Shutting down the Broker service")
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)
        os.exit(_exit_status)

    def SvcDoRun(self):
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_, ''))
        self.main()

    def main(self):
        try:
            of.broker.broker.start_broker()
        except Exception as e:
            write_to_event_log("Application", 1, "Error starting broker", str(e))


if __name__ == '__main__':
    win32serviceutil.HandleCommandLine(BrokerService)



""""
This modules holds the CumulativeDict class.
"""
import json

__author__ = "Nicklas Börjesson"


class CumulativeDict:
    """The CumulativeDicts class holds dictionaries indexed by names, adding one
    """

    """A dict namespaces, keys are namespaces"""
    _dicts = None

    """The default value of a new dict"""
    default = None

    def __init__(self, _default = None):
        self._dicts = {}
        self.default = _default

    def add_cumulatively(self, _dicts):
        """
        Add a definition to _dicts, _definition may
        :param _definition: The definition to add, may span several namaspaces
        :return:
        """

        def recurse_dict(_existing, _new):
            if isinstance(_new, dict):
                for _curr_new_key, _curr_new_value in _new.items():
                    if _curr_new_key in _existing and isinstance(_existing[_curr_new_key], dict) \
                            and isinstance(_existing[_curr_new_key], dict):
                        recurse_dict(_existing[_curr_new_key], _curr_new_value)
                    else:
                        _existing[_curr_new_key] = _curr_new_value

            else:
                raise Exception("Can only call cumulatively_add_definition with a dict")

        recurse_dict(self._dicts, _dicts)

    # Getters, setters and iterators

    def __getitem__(self, item):
        if item not in self._dicts:
            self._dicts[item] = dict(self.default)
        return self._dicts[item]

    def __setitem__(self, key, value):
        self._dicts[key] = value

    def __iter__(self):
        for _item in self._dicts:
            yield _item

    def values(self):
        return self._dicts.values()

    def keys(self):
        return self._dicts.keys()

    def items(self):
        return self._dicts.items()


    def load_dicts(self, _definition_files, _top_attribute=None):
        """
        Load namespaces from files
        :param _definition_files: A list of _definition_files
        """

        def load_definition_file(_filename):
            with open(_filename, "r") as _local_file:
                _local_def_data = json.load(_local_file)
                if _top_attribute is not None:
                    self.add_cumulatively(_local_def_data[_top_attribute])
                else:
                    self.add_cumulatively(_local_def_data)

        if _definition_files:
            for _curr_definition in _definition_files:
                load_definition_file(_curr_definition)


    def as_dict(self):
        return self._dicts



"""
This package holds the Optimal Framework broker, its libraries and UI
"""
__author__ = 'Nicklas Borjesson'

import runpy
def run_broker():
    runpy.run_module(mod_name="of.broker.broker", run_name="__main__")



import datetime
import hashlib

from of.broker.lib.schema_mongodb import mbe_object_id
from of.common.security.authentication import AuthenticationBackend


class MongoDBAuthBackend(AuthenticationBackend):
    # A db_access instance
    db_access = None

    def __init__(self, _db_access):
        self.db_access = _db_access


    def get_session(self, _session_id):
        _session_cond = {
            "conditions": {"_id": mbe_object_id(_session_id)},
            "collection": "session"
        }

        _sessions = self.db_access.find(_session_cond)
        if len(_sessions) > 0:
            if len(_sessions) > 1:
                raise Exception("session check: Multiple users returned by user query for " + _session_id)
            return _sessions[0]
        else:
            return None

    def get_user(self, _user_id):
        _user_condition = {
            "conditions": {"_id": mbe_object_id(_user_id), "schemaRef": "of://user.json"},
            "collection": "node"
        }
        _users = list(self.db_access.find(_user_condition))
        if len(_users) > 0:
            if len(_users) > 1:
                raise Exception("get user: Multiple users returned by user query for " + _user_id)
            return _users[0]
        else:
            return None

    def authenticate_username_password(self, _credentials):
        _cred_condition = {
                "conditions":
                        {
                            "credentials.usernamePassword.username": _credentials["usernamePassword"]["username"],
                         "credentials.usernamePassword.password": hashlib.md5(
                            _credentials["usernamePassword"]["password"].encode('utf-8')).hexdigest()

                        },
                "collection": "node"
            }

        # TODO: FTRT make a general credentials schema that is used by the user schema and use it to verify \
        # credential packets

        _users = self.db_access.find(_cred_condition)
        if len(_users) > 0:
            if len(_users) > 1:
                raise Exception("password_login error: Multiple users returned by user query.")

            _user = _users[0]
            # Create Session
            _session_data = {
                "createdWhen": str(datetime.datetime.utcnow()),
                "user_id": str(_user["_id"]),
                "schemaRef": "of://session.json"
            }

            _session_id = self.db_access.save(_session_data, _user)

            return _session_id, _user
        else:
            return None, None


    def logout(self, _session_id):

        _remove_cond = {
            "conditions": {"_id": mbe_object_id(_session_id)},
            "collection": "session"
        }

        self.db_access.remove_condition(_remove_cond, None)


"""
The utils module gathers all utility functions used in messaging
"""

import os
import socket
import platform
import datetime
import json
import logging


import requests
import sys


from requests.cookies import RequestsCookieJar

from of.common.logging import write_to_log, EC_NOTIFICATION, SEV_DEBUG, SEV_FATAL, EC_SERVICE, SEV_ERROR, \
    EC_COMMUNICATION
from of.common.messaging.factory import get_current_login
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning, SSLError

# Suppress insecure request
# TODO: However, this cho
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# TODO: get_python_versions should be duplicated into of.common.* (PROD-94)

__author__ = 'Nicklas Borjesson'


def make_log_prefix(_name):
    return _name + "(" + str(os.getpid()) + "):"

def sys_modules():
    result = []
    for curr_key in sys.modules.keys():
        result.append(curr_key)

    return result

def get_environment_data():
    def python_versions():
        _major, _minor, _release, _state, _build = sys.version_info
        return str(_major) + "." + str(_minor) + "." + str(_release) + " " + _state + " build " + str(_build)

    return {
        "hostname": socket.gethostname(),
        "implementation": {
            "language": "python",
            "version": python_versions(),
            "modules": sys_modules(),
        },
        "platform": platform.system(),
        "processor": platform.processor(),
        "systemPid": os.getpid(),
        "user": get_current_login(),
    }

def write_dbg_info(_data):
    write_to_log(_data, _category=EC_NOTIFICATION, _severity=SEV_DEBUG)

def register_at_broker(_address, _type, _server, _username, _password, _log_prefix="", _verify_SSL=True):
    _log_prefix = make_log_prefix(_log_prefix)

    _data = {
        "credentials": {
            "usernamePassword": {
                "username": _username,
                "password": _password
            }
        },
        "environment": get_environment_data(),
        "peerType": _type,
        "address": _address
    }
    # TODO: Credentials should not be outputted here in production.(ORG-112)
    write_dbg_info(_log_prefix + "[" + str(datetime.datetime.utcnow()) + "] Registering at broker API.")

    _headers = {'content-type': 'application/json'}
    _response = requests.post(_server + "/register", data=json.dumps(_data), auth=('user', 'pass'), headers=_headers,
                              verify=_verify_SSL)
    if _response.status_code == 500:
        write_dbg_info(_log_prefix + "Broker login failed with internal server error! Exiting.")
        return False
    if _response.status_code != 200:
        write_dbg_info(_log_prefix + "Broker login failed with error + "+ str(_response.status_code) + "! Exiting.")
        return False

    _response_dict = _response.json()
    if _response_dict is not None:
        _data = _response_dict

        if "session_id" in _data:
            write_dbg_info(_log_prefix + "Got a session id:" + _data["session_id"])
            return _data
        else:
            write_to_log(_log_prefix + "Broker login failed! Exiting.", _category=EC_SERVICE, _severity=SEV_ERROR)
        return False

    else:
        write_to_log(_log_prefix + "Broker login failed! Exiting.", _category=EC_SERVICE, _severity=SEV_ERROR)
        return False


def call_api(_url, _session_id, _data, _timeout=None, _print_log=None, _verify_SSL=True):
    """

    :param _url:
    :param _session_id:
    :param _data:
    :param _timeout:
    :param _print_log: Do not call write to log
    :return:
    """

    def do_log(_error, _category=EC_NOTIFICATION, _severity=SEV_DEBUG):
        if _print_log:
            print(_error)
        else:
            write_to_log(_data, _category=_category, _severity=_severity)
        return _error

    _cookie_jar = RequestsCookieJar()
    _cookie_jar.set(name="session_id", value=_session_id, secure=True)

    _headers = {'content-type': 'application/json'}

    _response = requests.post(_url, data=json.dumps(_data), headers=_headers, timeout=_timeout,
                              verify=_verify_SSL, cookies=_cookie_jar)

    _response_dict = None

    if _response.status_code != 200:
        do_log("Response code :" + str(_response.status_code))
        try:
            _response.raise_for_status()
        except Exception as e:
            raise Exception(do_log("Error in call_api:" + str(e), _category=EC_COMMUNICATION, _severity=SEV_ERROR))
    else:
        if _response.content:
            try:
                _response_dict = _response.json()
            except Exception as e:
                do_log("response.content didn't contain JSON data", _category=EC_COMMUNICATION, _severity=SEV_ERROR)
                _response_dict = None

    if _response_dict is not None:
        return _response_dict
    else:
        do_log("Got an empty response from server:" + str(_response.content), _category=EC_COMMUNICATION,
                     _severity=SEV_ERROR)

        return None


def make_error(_code, _message):
    return {
        "error_code": _code,
        "message": _message
    }


def message_is_none(_message, _property, _value):
    """
    If a property is unset, return value, otherwise the specified property

    :param _message: The message
    :param _property: The name of the property
    :param _value: The value
    :return: A value
    """
    if _property in _message:
        return _message[_property]
    else:
        return _value


class MultiprocessingLoggingHandler:
    # TODO: Should this be removed or is it used elsewhere?
    level = 0
    debug_prefix = None

    def handle(self, _record):
        write_dbg_info(self.debug_prefix + ": " + str(_record))

    def __init__(self, _log_prefix=None, _level=None):
        if _log_prefix is not None:
            self.debug_prefix = _log_prefix
        else:
            self.debug_prefix = "No debug_prefix set"

        if _level is not None:
            self.level = _level
        else:
            self.level = logging.INFO



"""
The handler module implements the Handler class.
"""

import os

from of.common.logging import write_to_log, EC_COMMUNICATION, SEV_DEBUG, EC_NOTIFICATION

__author__ = 'Nicklas Borjesson'


class Handler(object):
    """
    Handler is the base class for all queue item handlers.
    """

    #: The prefix for all log messages
    log_prefix = None
    #: The Optimal BPM process Id
    process_id = None

    #: The last message id, used to keep track of responses.
    _last_message_id = None
    # TODO: This is likely insufficient, look at new solution.

    def __init__(self, _process_id):
        self.process_id = _process_id
        self.log_prefix = str(os.getpid()) + "-" + self.__class__.__name__ + ": "

    def on_monitor_init(self, _monitor):
        """
        Override this to make something happen after the monitor has been initialized.

        :param _monitor: The monitor
        """
        pass

    def write_dbg_info(self, _data):
        """
        Shortcut to writing debug information
        :param _message: The message
        :param _severity: The severity of the log information, a constant defined in the built-in logging module
        """
        write_to_log(self.log_prefix + _data, _category=EC_NOTIFICATION, _severity=SEV_DEBUG, _process_id=self.process_id)


    def handle(self, _item):
        """
        This function is implemented by subclasses and is called when the monitor detects that an item has been queued.
        It is imperative that it doesn't contain synchronous code that might not return or otherwise hang in any way.
        Hint: Only use async I/0 and avoid complex library calls.

        :param _item: The item from the queue
        """
        raise Exception(self.log_prefix + "The required call handler method is not implemented.")

    def shut_down(self, _user_id):
        """
        Called by the monitor when shutting down. Override to provide own actions

        :param _user_id: The Optimal BPM user id

        """

        pass



"""
This process holds constants for the web socket error-codes
"""
__author__ = 'Nicklas Börjesson'

# Web socket error codes: http://tools.ietf.org/html/rfc6455#section-7.4.1

#: 1000 indicates a normal closure, meaning that the purpose for which the connection was established has been fulfilled.
WS_NORMAL_CLOSURE = 1000
#: 1001 indicates that an endpoint is "going away", such as a server going down or a browser having navigated away from a page.
GOING_AWAY = 1001
#: 1002 indicates that an endpoint is terminating the connection due to a protocol error.
PROTOCOL_ERROR = 1002
#: 1003 indicates that an endpoint is terminating the connection because it has received a type of data it cannot accept (e.g., an endpoint that understands only text data MAY send this if it receives a binary message).
UNACCEPTABLE_DATA = 1003
#: 1005 is a reserved value and MUST NOT be set as a status code in a Close control frame by an endpoint.  It is designated for use in applications expecting a status code to indicate that no status code was actually present."
NO_STATUS = 1005
#: 1006 is a reserved value and MUST NOT be set as a status code in a Close control frame by an endpoint.  It is designated for use in applications expecting a status code to indicate that the connection was closed abnormally, e.g., without sending or receiving a Close control frame.
ABNORMAL_CLOSE = 1006
#: 1007 indicates that an endpoint is terminating the connection because it has received data within a message that was not consistent with the type of the message (e.g., non-UTF-8 [RFC3629] data within a text message).
INCONSISTENT_DATA = 1007
#: 1008 indicates that an endpoint is terminating the connection because it has received a message that violates its policy.  This is a generic status code that can be returned when there is no other more suitable status code (e.g., 1003 or 1009) or if there is a need to hide specific details about the policy.
POLICY_VALIDATION = 1008
#: 1009 indicates that an endpoint is terminating the connection because it has received a message that is too big for it to process.
TOO_BIG = 1009
#: 1010 indicates that an endpoint (client) is terminating the connection because it has expected the server to negotiate one or more extension, but the server didn't return them in the response message of the WebSocket handshake.  The list of extensions that are needed SHOULD appear in the /reason/ part of the Close frame. Note that this status code is not used by the server, because it can fail the WebSocket handshake instead.
BAD_SERVER_RESPONSE = 1010
#: 1011 indicates that a server is terminating the connection because it encountered an unexpected condition that prevented it from fulfilling the request.
UNEXPECTED_CONDITION = 1011
#: 1015 is a reserved value and MUST NOT be set as a status code in a Close control frame by an endpoint. It is designated for use in applications expecting a status code to indicate that the connection was closed due to a failure to perform a TLS handshake (e.g., the server certificate can't be verified).
TLS_FAILURE = 1015

#: 4011 is a private value that indicates that a socket is closed because the broker is restarting, this to tell the agents to wait a short while and re-register and reconnect
BROKER_RESTARTING = 4011
#: 4012 is a private value that indicates that a socket is closed because the broker is shutting down, this to tell agents to wait a longer and periodically reconnect
BROKER_SHUTTING_DOWN = 4012


#: 1004 is Reserved.  The specific meaning might be defined in the future.



"""
This package holds the Optimal Framework source package
"""
__author__ = 'Nicklas Borjesson'

# Version and release information used by Sphinx for documentation and setuptools for package generation.
__version__ = '0.9'
__release__ = '0.9.2'
__copyright__ = '2010-2016, Optimal BPM'




"""
The constants module holds constants for all the schemas defined in Optimal Framework
"""
__author__ = 'Nicklas Borjesson'

"""
Constants for _ids in the Optimal Framework database
TODO: These are only constant for this instance of the database (perhaps XOR-them together with environment Id?)
"""
# Hard coded locations of always present nodes

id_administration = "000000010000010001e64c23"
id_groups = "000000010000010001e64c24"
id_users = "000000010000010001e64c25"
id_rights = "000000010000010001e64c26"
id_right_admin_everything = "000000010000010001e64c27"
id_group_administrators = "000000010000010001e64c28"
id_group_users = "000000010000010001e64c29"

id_users = "000000010000010001e64c25"
id_peers = "000000010000010002e64d03"

# Shorthand for a valid 24-digit zero object id
zero_object_id = "000000000000000000000000"

"""
Schema definition constants

Note: A new schema must be both given schemaRef and be added to category
"""



# Schema category map. Used by handlers.
schema_categories = {
    # Message category - Messages between peers - forwarded by the broker to the destination if not itself
    "of://message.json": "message",
    "of://message_error.json": "message",
    # Node category - Nodes in the node tree, cannot be messages
    "of://node_broker.json": "node",
    "of://node_admin.json": "node",
    "of://node_peer.json": "node",
    # Log category - Log messages - written to the log collection by the broker
    "of://log_progression.json": "log",
    "of://log_process_state.json": "log",
    "of://event.json": "log",
    # Process category - Process instance data - written to the process collection by the broker
    "of://process_system.json": "process"
    # Control category - Control messages for runtime entities
    # None for broker yet.
}

# These messages should are intercepted and stored by the broker
intercept_schema_ids = []

peer_type__schema_id = {

    "broker": "of://node_broker.json",
    "admin": "of://node_admin.json",
    "peer": "of://node_peer.json"
}


def peer_type_to_schema_id(_peer_type):
    if _peer_type in peer_type__schema_id:
        return peer_type__schema_id[_peer_type]
    else:
        raise Exception("Invalid peer type:" + _peer_type)



"""
Created on Mar 6, 2015

The groups module contains functionality for checking group level rights.

@author: Nicklas Boerjesson


"""

from decorator import getfullargspec

__author__ = 'nibo'

_groups = None


def init_groups(_database_access):
    """
    Initiate the global groups object, load all groups from the provided database

    :param _database_access: A database access object
    :return: The global group object

    """

    global _groups
    _groups = Groups(_database_access)


def has_right(_right, _user):
    """

    Global convenience function, see Groups.has_right

    :param _right:

    """
    global _groups
    if _groups:
        _groups.has_right(_right, _user)


def user_in_any_of_groups(_user, _groups):
    """
    Check if a user belongs to any of the groups
    :param _user: A user instance
    :param _groups: A list of groups to check
    :return: True if in any of the groups
    """
    _user_groups = _user["groups"]
    for _group in _groups:
        if _group in _user_groups:
            return True
    return False

def aop_has_right(_rights):
    """

    A decorator to supply the has_right functionality. Takes a function pointer.
    It is necessary to make it able to get the data in runtime.
    TODO: That is quite ugly.

    :param _rights_function: A a list with rights or a pointer to a function that returns the right to check for
    :return: returns a decorator function.


    """

    def wrapper(func):
        def wrapped_f(*args, **kwargs):
            # Get a list of the names of the non-keyword arguments
            if "_user" in kwargs:
                _user = kwargs["_user"]
            else:

                # TODO: Add support for keyword arguments, can be convenient in cases where code is called from python
                _argument_specifications = getfullargspec(func)
                try:
                    # Try and find _session_id
                    user_idx = _argument_specifications.args.index("_user")
                except:
                    raise Exception("Has right aspect for \"" + func.__name__ + "\": No _user parameter in function, internal error.")

                if user_idx > len(args) - 1:
                    raise Exception(
                        "Has right aspect for \"" + func.__name__ + "\": The _user parameter isn't supplied.")


                _user = args[user_idx]

            if isinstance(_rights, list):
                has_right(_rights, _user)
            elif callable(_rights):
                has_right(_rights(), _user)




            return func(*args, **kwargs)

        return wrapped_f

    return wrapper


class RightCheckError(Exception):
    """Exception class for Rights errors"""
    pass


class Groups():
    """
    The Groups class is the central location for group management in MBE and holds the group cache.

    """
    _groups = {}

    def __init__(self, _database_access):
        """
        The class initiates by loading all groups from the database.

        :param _database_access: A database access object instance


        """
        self._groups = {}
        self._database_access = _database_access
        self.reload_all()

    def reload_all(self):
        """
        Reload all groups from database
        """
        self._groups.clear()
        _cursor = self._database_access.find(
            {"collection": "node", "conditions": {"schemaRef": "of://group.json"}})
        for _curr_group in _cursor:
            self._groups[_curr_group["_id"]] = _curr_group

    def has_right(self, _right, _user):
        """
        Check if a certain user, through group membership, holds a certain right.
        Raises a RightCheckError if npt

        :param _right: The right in question
        :param _user: A user object

        """

        if type(_right) is list:
            for _curr_group in _user["groups"]:
                for _curr_right in _right:
                    if _curr_right in self._groups[_curr_group]["rights"]:
                        return _curr_group
        else:
            for _curr_group in _user["groups"]:
                if _right in self._groups[_curr_group]["rights"]:
                    return _curr_group

        # TODO: Add nice-looking alert for all errors in UI
        _error = "The user \"" + _user["name"] + "\" doesn't have the " + str(_right) + " right."
        self._database_access.logging.log_security("right",
                                                   _error,
                                                   _user["_id"], None)

        raise RightCheckError("The user \"" + _user["name"] + "\" doesn't have the " + str(_right) + " right.")

    def __iter__(self):
        """Access function for making the class iterable."""
        return self._groups

    def __getitem__(self, item):
        """Access function for making the class iterable."""
        return self._groups[item]



"""
The init module is responsible for initializing an MBE backend database.
"""

import os

__author__ = 'Nicklas Boerjesson'

import json

from pymongo.mongo_client import MongoClient

from of.broker.lib.access import DatabaseAccess

script_dir = os.path.dirname(__file__)


def get_empty_db_da(_database_name, _json_schema_folders=None, _uri_handlers = None):
    """
    Create an empty database. Drops any existing.

    :param _database_name: The name of the database
    :return: A database access object for the database
    :param _json_schema_folders: A list of application specific JSON schema folders

    """
    _client = MongoClient()
    if _database_name in _client.database_names():
        _client.drop_database(_client[_database_name])
    _database = _client[_database_name]
    return DatabaseAccess(_database=_database, _json_schema_folders=_json_schema_folders, _uri_handlers=_uri_handlers)


def _import_init_file(_filename, _da):
    """
    Imports the content of the file to tha database

    :param _filename: File to import
    :param _da: A data access instance
    :return:

    """
    json_data = open(_filename)
    _data = json.load(json_data)
    for _curr_data in _data:
        print("init - " + _curr_data["name"])
        _da.save(_curr_data, _user=None, _allow_save_id=True)
    json_data.close()


def init_database(_database_name="test_database", _data_files=None, _json_schema_folders=None, _uri_handlers= None):
    """
    Initialises a MBE database. Drops any existing databases and import the provided files

    :param _database_name: The name of the database.
    :param _data_files: A list of application specific additional files to import
    :param _json_schema_folders: A list of application specific JSON schema folders
    :return: A database access object for the database

    """
    if not _data_files:
        _data_files = []

    _da = get_empty_db_da(_database_name, _json_schema_folders = _json_schema_folders, _uri_handlers= _uri_handlers)
    # Init by loading the MBE base structure
    _import_init_file(os.path.join(script_dir, 'init.json'), _da)

    # Import the rest of the data
    for _filename in _data_files:
        _import_init_file(_filename, _da)

    return _da


if __name__ == '__main__':
    init_database()




"""
    The CherryPy-module provides a web server plugin for CherryPy.
"""

from http.cookies import SimpleCookie

import cherrypy
from decorator import decorator

from of.broker.lib.node import Node
from of.common.security.authentication import check_session, login, AuthenticationError, logout

__author__ = 'Nicklas Borjesson'


def _check_session_aspect(func, *args, **kwargs):
    """CherryPy-specific check_session aspect code, checks if there is a valid cookie"""

    try:
        # Try and find _session_id
        _session_id = cherrypy.request.cookie["session_id"].value
    except:
        raise AuthenticationError("Authentication aspect for \"" + func.__name__ + "\": No session_id cookie")

    # Set user variable and execute function.
    kwargs["_user"] = check_session(_session_id)
    kwargs["_session_id"] = _session_id
    return func(*args, **kwargs)


def aop_check_session(f):
    """
    CherryPy-specific session checking decorator.

    :param f: the function

    """
    return decorator(_check_session_aspect, f)


def _login_json_aspect(func, *args, **kwargs):
    """
    CherryPy-specific session login aspect.

    """
    _message = cherrypy.request.json

    if _message is not None and isinstance(_message, dict):
        # TODO: Add a hook for custom login

        if "credentials" not in _message:
            raise AuthenticationError("_login_json_aspect for \"" + func.__name__ + "\": No credentials in request")

        # Already logged in, log out old session
        if "session_id" in cherrypy.request.cookie and cherrypy.request.cookie["session_id"].value != "":
            print("_login_json_aspect: Logging out old session: " + cherrypy.request.cookie["session_id"].value)
            logout(cherrypy.request.cookie["session_id"].value)


        _session_id, _user = login(_message["credentials"])
        _session_cookie = cherrypy.response.cookie
        _session_cookie["session_id"] = _session_id
        _session_cookie["session_id"]["secure"] = True
        _session_cookie["session_id"]["path"] = "/"

        kwargs["_message"] = _message
        kwargs["_session_id"] = _session_id
        kwargs["_user"] = _user
        return func(*args, **kwargs)
    else:
        raise AuthenticationError("Authentication aspect for \"" + func.__name__ + "\": No message structure")


def aop_login_json(f):
    """
    CherryPy-specific session login decorator.

    """
    return decorator(_login_json_aspect, f)


class CherryPyNode(object):
    """
    The CherryPyNode class is a plugin for CherryPy to expose the MBE node-functionality.
    Each function maps to counterparts in the node class.

    """
    # A local instance of the MBE Node API
    _node = None

    def __init__(self, _database_access):
        self._node = Node(_database_access)

    # noinspection PyUnusedLocal
    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_login_json
    def login(self, **kwargs):
        return {}

    # noinspection PyUnusedLocal
    @cherrypy.expose
    @cherrypy.tools.json_out(content_type='application/json')
    def logout(self, **kwargs):
        if "session_id" in cherrypy.request.cookie:
            logout(cherrypy.request.cookie["session_id"].value)
            cherrypy.response.cookie["session_id"] = SimpleCookie()
            cherrypy.response.cookie["session_id"]['expires'] = 0
        return {}

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def save(self, **kwargs):
        return self._node.save(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def find(self, **kwargs):
        return self._node.find(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def lookup(self, **kwargs):
        return self._node.lookup(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def remove(self, **kwargs):
        return self._node.remove(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def load_node(self, **kwargs):
        return self._node.load_node(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def load_children(self, **kwargs):
        return self._node.load_children(cherrypy.request.json, kwargs["_user"])

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def history(self, **kwargs):
        return self._node.history(cherrypy.request.json, kwargs["_user"])



    @cherrypy.expose
    @cherrypy.tools.json_out(content_type='application/json')
    @aop_check_session
    def get_schemas(self, **kwargs):
        return self._node.get_schemas(kwargs["_user"])



import time

import cherrypy
from ws4py.websocket import WebSocket

from of.common.messaging.websocket import OptimalWebSocket
import of.common.messaging.websocket
__author__ = 'Nicklas Borjesson'


class BrokerWebSocket(OptimalWebSocket, WebSocket):
    """
    This class holds a connection from a web socket client
    """
    monitor_message_queue_thread = None

    def __init__(self,  sock, protocols=None, extensions=None, environ=None, heartbeat_freq=None):
        self.log_prefix = ""
        self.write_dbg_info("BrokerWebSocket: Peer connected:" + str(cherrypy.request.remote.ip))
        super(BrokerWebSocket, self).__init__(sock, protocols, extensions, environ, heartbeat_freq)

        self.write_dbg_info("BrokerWebSocket: New peer session, init " + str(cherrypy.request.cookie['session_id'].value))
        self.init(_session_id=cherrypy.request.cookie['session_id'].value)

    def close(self, code=1000, reason=''):
        """
        Closes the connection, and unregisters the web socket at the message monitor
        :param code: A web socket error code as defined in optimalbpm.common messaging.constants, information at: http://tools.ietf.org/html/rfc6455#section-7.4.1
        :param reason: A string describing the reason for closing the connection
        """
        self.write_dbg_info(self.log_prefix + "Told to close (session_id:" + str(self.session_id) + ") , code: " + str(code) + ", reason: " + str(reason))
        if self.monitor_message_queue_thread:
            self.monitor_message_queue_thread.terminated = True

        of.common.messaging.websocket.monitor.handler.unregister_web_socket(self)

        super(BrokerWebSocket, self).close(code, reason)

        self.write_dbg_info("Web socket for "+ self.address + " closed.")

    def closed(self, code, reason=None):
        """
        Called when the socket is closed.
        :param code: A web socket error code as defined in: http://tools.ietf.org/html/rfc6455#section-7.4.1
        :param reason: A string describing the reason for closing the connection
        """
        self.write_dbg_info("Closed, code: " + str(code) + ", reason: " + str(reason))

class MockupWebSocket(BrokerWebSocket):
    """
    This is a mockup class to only test the message broker itself.
    """

    on_message = None
    context = None

    def __init__(self, session_id=None, context=None):
        """
        Any changes made here must be reflected in BrokerWebSocket.__init__
        """

        class MockupSocket:
            """ A socket implementation that does *nothing* """

            def __init__(self):
                self.server_terminated = False

            def close(self):
                print("MockupSocket - Told to close.")

            def sendall(self, b):
                pass

        super(BrokerWebSocket, self).__init__(sock=MockupSocket(), protocols=None, extensions=None, environ=None,
                                               heartbeat_freq=None)

        self.session_id = session_id
        self.context = context

        self.init(session_id)

    def received_message(self, message):
        """
        Receives a message from a peer and puts it on the queue
        """
        self.received_last_message_at = time.perf_counter()
        super(MockupWebSocket, self).received_message(message)

    def send_message(self, message):
        """
            Replaces the BrokerWebSocket.send_message
        """
        self.message = message
        self.sent_last_message_at = time.perf_counter()
        if self.on_message:
            self.on_message(self, message)
        print(self.log_prefix + "got a message to send to its peer from " + message["source"])



"""
The init_env module/script initializes an Optimal Framework testing environment.
"""
import os
import sys

from bson.objectid import ObjectId

script_dir = os.path.dirname(__file__)
# Add relative optimal bpm path to be able to load the modules of this repository properly
sys.path.append(os.path.join(script_dir, "../../../"))

from of.broker.lib.auth_backend import MongoDBAuthBackend
from of.common.security.authentication import init_authentication
from of.broker.testing.init import init_database
from of.schemas.validation import of_uri_handler

__author__ = 'nibo'




def init_env(_database_name = "test_of", _context=None, _data_files=[], _json_schema_folders=[], _uri_handlers={}):
    """
    Initiates the test_broker database
    :param _context: If set, logs in and adds db_access, auth, session_id and peer_process_id properties
    :return:
    """
    _data_files += [os.path.join(script_dir, "data_struct.json")]
    _json_schema_folders += [os.path.abspath(os.path.join(script_dir, "..", "..", "schemas"))]
    _uri_handlers.update({"of": of_uri_handler})

    _db_access = init_database(_database_name, _data_files=_data_files,
                               _json_schema_folders=_json_schema_folders,
                               _uri_handlers=_uri_handlers)

    if _context:
        _context.db_access = _db_access
        _context.auth = init_authentication(MongoDBAuthBackend(_context.db_access))

        _context.session_id, _context.user = _context.auth.login(
            {"usernamePassword": {"username": "tester", "password": "test"}})

        _context.peer_process_id = str(ObjectId())


if __name__ == "__main__":
    init_env()



"""
    The node module provides the node-API for OF. All interaction with the node-collection should happen
    through the Node-class

    Note: Sphinx does not properly document this module with the decorators, comment them before running sphinx
"""

from bson.objectid import ObjectId

from of.broker.lib.auth_backend import MongoDBAuthBackend
from of.common.security.authentication import init_authentication
from of.common.security.groups import aop_has_right, init_groups

from of.broker.lib.schema_mongodb import mbe_object_id
from of.common.security.permission import filter_by_group



# TODO: Separate or classify all MongoDB-specific querying
from of.schemas.constants import id_right_admin_everything

__author__ = 'nibo'

node_rights = None
"""
A cache of node rights
"""


def sanitize_node(_node):
    """
    Sanitize a node instance so that sensitive data isn't disseminated
    :param _node: The node to sanitize
    :return:
    """
    if "credentials" in _node:
        _node.pop("credentials")
    if "canRead" in _node:
        _node.pop("canRead")
    if "canWrite" in _node:
        _node.pop("canWrite")

    return _node

def get_node_rights():
    """
    This function is necessary to hand runtime data to the decorators

    :return: the node rights

    """
    return node_rights


class Node():
    """
    The node class is used for all node management in OF.
    """
    database_access = None
    rights = None

    def __init__(self, _database_access, _rights=None):
        """
        Initializes that node object and groups.

        :param _database_access: A database access instance

        """
        if _database_access is None:
            raise Exception("Node: The database access parameter is None, it must be set for the Node class to work")

        self.database_access = _database_access
        global node_rights

        if _rights is not None:
            node_rights = [id_right_admin_everything] + _rights
        else:
            node_rights = [id_right_admin_everything]

        init_groups(self.database_access)

    def _string_to_object_ids(self, _data):

        if isinstance(_data, list):
            _destination = []
            for _curr_row in _data:
                _destination.append(self._string_to_object_ids(_curr_row))

            return _destination
        elif isinstance(_data, dict):
            _destination = {}
            for _curr_key, _curr_value in _data.items():
                _destination[_curr_key] = self._string_to_object_ids(_curr_value)
            return _destination

        elif _data is not None and type(_data) != ObjectId and _data[:9] == "ObjectId(":
            return ObjectId(_data[9:-1])
        else:
            return _data

    @aop_has_right(get_node_rights)
    def save(self, _document, _user):
        """save(self, _document, _user)

        Saves an MBE node descendant to the database.

        :param _document: The MBE node document to save.
        :param _user: The current user
        :return: A structure detailing the save result

        """

        # The data must have a schema id key and it must say that the collection is "node".
        self.database_access.verify_document(_document, "node", "Node.save")

        if "_id" in _document:  # It is an existing node, check if user can write to it.
            _old_document_result = filter_by_group(
                self.database_access.find(
                    {"conditions": {"_id": mbe_object_id(_document["_id"])}, "collection": "node"}),
                "canWrite", _user, "Node.save(existing node)")
            if len(_old_document_result) == 1:
                _old_document = _old_document_result[0]
            else:
                raise Exception("Needed exactly one(1) existing node, found " + str(len(_old_document_result)) +
                                ". _id: " + _document["_id"])

        elif "parent_id" in _document:  # It is a new node, check if the parent allows user to create it
            filter_by_group(
                self.database_access.find(
                    {"conditions": {"_id": mbe_object_id(_document["parent_id"])}, "collection": "node"}),
                "canWrite", _user, "Node.save(new node)")
            _old_document = None
        else:
            raise Exception("Node.save: Data must have a parentId or an _id.")

        # TODO: Specify how the result actually looks
        return self.database_access.save(_document, _user, _old_document)

    @aop_has_right(get_node_rights)
    def find(self, _conditions, _user, _error_prefix_if_not_allowed=None):
        """find(_conditions,_user)
        Returns a list of nodes matching the MBE condition.
        Note: This method is literate, it takes ObjectId instances when comparing to ObjectId instances.
        But JSON doesn't handle ObjectId. So instead it supports an ObjectId(000000010000010001ee3214)-syntax
        and when it encounters those values, it converts them into ObjectId instances.

        :param _conditions: A condition
        :param _user: A user object
        :return: A list of nodes


        """


        # Filter result by canRead groups
        return filter_by_group(self.database_access.find(
            {"conditions": self._string_to_object_ids(_conditions), "collection": "node"}),
            "canRead", _user, self.database_access, _error_prefix_if_not_allowed=_error_prefix_if_not_allowed)

    @aop_has_right(get_node_rights)
    def load_children(self, _parent_id, _user):
        """
        Returns a list of child nodes whose _parent_ids match _parent_id

        :param _parent_id: An object with a parent_id property
        :param _user: A user object
        :return: A list of nodes

        """

        # Filter result by canRead groups
        return filter_by_group(
            self.database_access.find(
                {"conditions": {"parent_id": mbe_object_id(_parent_id["parent_id"])}, "collection": "node"}),
            "canRead", _user, self.database_access)

    @aop_has_right(get_node_rights)
    def load_node(self, _id, _user):
        """
        Returns the child with _id.

        :param _id: An object with a id property
        :param _user: A user object
        :return: A list of nodes

        """

        # Filter result by canRead groups
        return filter_by_group(self.database_access.find(
            {"conditions": {"_id": mbe_object_id(_id["_id"])}, "collection": "node"}),
            "canRead", _user, self.database_access,
            _error_prefix_if_not_allowed="Node.load_node: Not permissioned to load this node. _id: " + _id["_id"])[0]

    @aop_has_right(get_node_rights)
    def lookup(self, _conditions, _user):
        """
        Returns a list of value-name objects for use in drop downs.
        As find, handles the ObjectId(000000010000010001ee3214) string format. (no quotes)

        :param _conditions: An MBE condition
        :param _user: A user object
        :return: A list of lookup objects.

        """

        # Only allow node collection, but allow not specifying schema in conditions.
        self.database_access.verify_condition(_conditions, "node", "Node.find")

        print(str(_conditions))
        # Filter result by canRead groups
        return [{"value": _row["_id"], "text": _row["name"]} for _row in
                filter_by_group(self.database_access.find(self._string_to_object_ids(_conditions)), "canRead", _user,
                                self.database_access)]

    @aop_has_right(get_node_rights)
    def remove(self, _id, _user):
        """
        Remove the specified node

        :param _id: An object with a node _id field
        :param _user: A user object
        :return: Result of removal

        """

        # Recurse and gather all children

        def _recurse_children(_parent_node):
            _result = [_parent_node]
            _child_nodes = self.database_access.find(
                {"conditions": {"parent_id": _parent_node["_id"]}, "collection": "node"},
                _do_not_fix_object_ids=True)
            for _curr_child in _child_nodes:
                _result += _recurse_children(_curr_child)

            return _result

        # Load the current node
        _node = self.database_access.find({"conditions": {"_id": mbe_object_id(_id["_id"])}, "collection": "node"},
                                          _do_not_fix_object_ids=True)[0]

        _all_nodes = _recurse_children(_node)

        # Check permissions, as the remove_documents function works with documents directly the from database,
        # use objectIds.
        _documents = filter_by_group(
            _all_nodes,
            "canWrite", _user, self.database_access,
            _error_prefix_if_not_allowed="Node.remove: Missing permissions to remove node or sub nodes.",
            _use_object_id=True)

        # TODO: Specify removal result structure
        return self.database_access.remove_documents(_documents, _user, "node", )

    @aop_has_right(get_node_rights)
    def history(self, _id, _user):
        """

        Return the change history for the specified node.

        :param _id: An object with a node _id field
        :param _user: A user object
        :return:

        """
        object_id = mbe_object_id(_id["_id"])
        # Filter result by canRead
        filter_by_group(self.database_access.find(
            {"conditions": {"_id": object_id}, "collection": "node"}), "canRead",
            _user, self.database_access,
            _error_prefix_if_not_allowed="Node.history: No permissions to view details of this node. _id: " +
            _id["_id"])

        return self.database_access.find({"conditions": {"node_id": object_id}, "collection": "log"})


    def get_schemas(self, _user):
        print("Request for a list of schemas")
        return self.database_access.schema_tools.json_schema_objects


"""
This module contains the Optimal BPM Repositories API
"""

import os
from bson.objectid import ObjectId


__author__ = 'nibo'

from dulwich.repo import Repo

class Repositories():
    """
        Repositories
    """

    def __init__(self, _db_access, _node):

        self.node = _node

    def refresh_cache(self):
        """

        :return:
        """
        pass



    def get_hash(self, _id, _user):
        """
        Returns the hashes
        :param _id: The _id of the repository
        :return: The hash of the repository
        """
        _result = self.node.find({"_id": ObjectId(_id)}, _user, _error_prefix_if_not_allowed = "This user doesn't have permissions for this repository.")
        _repo = Repo(os.path.join(_result["folder"], str(_id)))
        return _repo.head()


    def init_repository(self, _id, _folder=None):
        """
        Initialize a repository
        :param _id: repository_id
        :param _folder:
        :return:
        """
        if _folder is None:
            _repo_folder = os.path.join(os.path.expanduser(self.default_folder), str(_id))
        else:
            _repo_folder = os.path.join(os.path.expanduser(_folder), str(_id))

        from os import mkdir
        mkdir(_repo_folder)
        # TODO: Implement repositories (PROD-11)
        #_repo = Repo.init(_repo_folder)

        # return _repo.head()